@@ -3,7 +3,7 @@
 # Copyright (c) 1999 by Megginson Technologies.
 # No warranty.  Commercial and non-commercial use freely permitted.
 #
-# $Id: Writer.pm,v 1.2 2003/10/18 20:18:58 ed Exp $
+# $Id: Writer.pm,v 1.6 2004/02/22 16:09:16 ed Exp $
 ########################################################################
 
 package XML::Writer;
@@ -14,7 +14,7 @@ use strict;
 use vars qw($VERSION);
 use Carp;
 use IO::Handle;
-$VERSION = "0.4.1";
+$VERSION = "0.4.2";
 
 
 
@@ -322,7 +322,30 @@ sub new {
     }
   };
 
-  
+  my $raw = sub {
+    $output->print($_[0]);
+    # Don't set $hasData or any other information: we know nothing
+    # about what was just written.
+    #
+  };
+
+  my $cdata = sub {
+      my $data = $_[0];
+      $data    =~ s/\]\]>/\]\]\]\]><!\[CDATA\[>/g;
+      $output->print("<![CDATA[$data]]>");
+      $hasData = 1;
+  };
+
+  my $SAFE_cdata = sub {
+    if ($elementLevel < 1) {
+      croak("Attempt to insert characters outside of document element");
+    } elsif ($dataMode && $hasElement) {
+      croak("Mixed content not allowed in data mode: characters");
+    } else {
+      &{$cdata};
+    }
+  };
+
 				# Assign the correct closures based on
 				# the UNSAFE parameter
   if ($unsafe) {
@@ -334,7 +357,10 @@ sub new {
 	     'STARTTAG' => $startTag,
 	     'EMPTYTAG' => $emptyTag,
 	     'ENDTAG' => $endTag,
-	     'CHARACTERS' => $characters};
+	     'CHARACTERS' => $characters,
+	     'RAW' => $raw,
+	     'CDATA' => $cdata,
+	    };
   } else {
     $self = {'END' => $SAFE_end,
 	     'XMLDECL' => $SAFE_xmlDecl,
@@ -344,7 +370,10 @@ sub new {
 	     'STARTTAG' => $SAFE_startTag,
 	     'EMPTYTAG' => $SAFE_emptyTag,
 	     'ENDTAG' => $SAFE_endTag,
-	     'CHARACTERS' => $SAFE_characters};
+	     'CHARACTERS' => $SAFE_characters,
+	     'RAW' => $raw,                    # deliberately unsafe
+	     'CDATA' => $SAFE_cdata,
+	    };
   }
 
 				# Query methods
@@ -491,6 +520,16 @@ sub dataElement {
   $self->endTag($name);
 }
 
+#
+# Write a simple CDATA element.
+#
+sub cdataElement {
+    my ($self, $name, $data, %atts) = (@_);
+    $self->startTag($name, %atts);
+    $self->cdata($data);
+    $self->endTag($name);
+}
+
 #
 # Write character data.
 #
@@ -499,6 +538,22 @@ sub characters {
   &{$self->{CHARACTERS}};
 }
 
+#
+# Write raw, unquoted, completely unchecked character data.
+#
+sub raw {
+  my $self = shift;
+  &{$self->{RAW}};
+}
+
+#
+# Write CDATA.
+#
+sub cdata {
+    my $self = shift;
+    &{$self->{CDATA}};
+}
+
 #
 # Query the current element.
 #
@@ -950,7 +1005,7 @@ XML::Writer - Perl extension for writing XML documents.
 =head1 SYNOPSIS
 
   use XML::Writer;
-  use IO;
+  use IO::File;
 
   my $output = new IO::File(">output.xml");
 
@@ -1184,6 +1239,29 @@ You may invoke this method only within the document element
 In data mode, you must not use this method to add whitespace between
 elements.
 
+=item raw($data)
+
+Print data completely unquoted and unchecked to the XML document.  For
+example C<raw('<')> will print a literal < character.  This
+necessarily bypasses all well-formedness checking, whether in unsafe
+mode or not.
+
+This can sometimes be useful for printing entities which are defined
+for your XML format but the module doesn't know about, for example
+&nbsp; for XHTML.
+
+=item cdata($data)
+
+As C<characters()> but writes the data quoted in a CDATA section, that
+is, between <![CDATA[ and ]]>.  If the data to be written itself
+contains ]]>, it will be written as several consecutive CDATA
+sections.
+
+=item cdataElement($name, $data [, $aname1 => $value1, ...])
+
+As C<dataElement()> but the element content is written as one or more
+CDATA sections (see C<cdata()>).
+
 =item setOutput($output)
 
 Set the current output destination, as in the OUTPUT parameter for the
